use std::collections::HashSet;

use proc_macro2::{Ident, Span};

use btrun::End;

use crate::{
    error::{self, MaybeErr},
    model::ProtocolModel,
    parsing::{DestinationState, GetSpan, Message, State},
};

impl ProtocolModel {
    #[allow(dead_code)]
    pub(crate) fn validate(&self) -> syn::Result<()> {
        self.all_states_declared_and_used()
            .combine(self.receivers_and_senders_matched())
            .combine(self.no_undeliverable_msgs())
            .combine(self.replies_expected())
            .combine(self.clients_only_receive_replies())
            .into()
    }

    /// Verifies that every state which is declared is actually used.
    fn all_states_declared_and_used(&self) -> MaybeErr {
        let end = Ident::new(End::ident(), Span::call_site());
        let mut declared: HashSet<&Ident> = HashSet::new();
        declared.insert(&end);
        for actor_def in self.def().actor_defs.iter() {
            for state in actor_def.states.as_ref().iter() {
                declared.insert(state);
            }
        }
        let mut used: HashSet<&Ident> = HashSet::with_capacity(declared.len());
        for transition in self.def().transitions.iter() {
            let in_state = &transition.in_state;
            used.insert(&in_state.state_trait);
            used.extend(
                in_state
                    .owned_states
                    .as_ref()
                    .iter()
                    .map(|ident| ident.as_ref()),
            );
            if let Some(in_msg) = transition.in_msg() {
                used.extend(
                    in_msg
                        .owned_states
                        .as_ref()
                        .iter()
                        .map(|ident| ident.as_ref()),
                );
            }
            for out_states in transition.out_states.as_ref().iter() {
                used.insert(&out_states.state_trait);
                used.extend(
                    out_states
                        .owned_states
                        .as_ref()
                        .iter()
                        .map(|ident| ident.as_ref()),
                );
            }
            // We don't have to check the states referred to in out_msgs because the
            // receivers_and_senders_matched method ensures that each of these exists in a receiver
            // position.
        }
        let undeclared: MaybeErr = used
            .difference(&declared)
            .map(|ident| syn::Error::new(ident.span(), error::msgs::UNDECLARED_STATE_ERR))
            .collect();
        let unused: MaybeErr = declared
            .difference(&used)
            .filter(|ident| **ident != End::ident())
            .map(|ident| syn::Error::new(ident.span(), error::msgs::UNUSED_STATE_ERR))
            .collect();
        undeclared.combine(unused)
    }

    /// Ensures that the recipient state for every sent message has a receiving transition
    /// defined, and every receiver has a sender. Note that each message isn't required to have a
    /// unique sender or a unique receiver, just that at least one of each much be defined.
    fn receivers_and_senders_matched(&self) -> MaybeErr {
        let mut senders: HashSet<(&State, &Message)> = HashSet::new();
        let mut receivers: HashSet<(&State, &Message)> = HashSet::new();
        for transition in self.def().transitions.iter() {
            if let Some(msg) = transition.in_msg() {
                receivers.insert((&transition.in_state, msg));
            }
            for dest in transition.out_msgs.as_ref().iter() {
                let dest_state = match &dest.state {
                    DestinationState::Individual(dest_state) => dest_state,
                    DestinationState::Service(dest_state) => dest_state,
                };
                senders.insert((dest_state, &dest.msg));
            }
        }
        let extra_senders: MaybeErr = senders
            .difference(&receivers)
            .map(|pair| syn::Error::new(pair.1.msg_type.span(), error::msgs::UNMATCHED_SENDER_ERR))
            .collect();
        let extra_receivers: MaybeErr = receivers
            .difference(&senders)
            .map(|pair| {
                syn::Error::new(pair.1.msg_type.span(), error::msgs::UNMATCHED_RECEIVER_ERR)
            })
            .collect();
        extra_senders.combine(extra_receivers)
    }

    /// Checks that messages are only sent to destinations which are either services, states
    /// which are owned by the sender, listed in the output states, or that the message is a
    /// reply.
    fn no_undeliverable_msgs(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.def().transitions.iter() {
            let mut allowed_states: Option<HashSet<&Ident>> = None;
            for dest in transition.out_msgs.as_ref().iter() {
                if dest.msg.is_reply() {
                    continue;
                }
                match &dest.state {
                    DestinationState::Service(_) => continue,
                    DestinationState::Individual(dest_state) => {
                        let allowed = allowed_states.get_or_insert_with(|| {
                            transition
                                .out_states
                                .as_ref()
                                .iter()
                                .map(|state| state.state_trait.as_ref())
                                .chain(
                                    transition
                                        .in_state
                                        .owned_states
                                        .as_ref()
                                        .iter()
                                        .map(|ident| ident.as_ref()),
                                )
                                .collect()
                        });
                        if !allowed.contains(dest_state.state_trait.as_ref()) {
                            err = err.combine(
                                syn::Error::new(
                                    dest_state.state_trait.span(),
                                    error::msgs::UNDELIVERABLE_ERR,
                                )
                                .into(),
                            );
                        }
                    }
                }
            }
        }
        err
    }

    /// Verifies that exactly one reply is sent in response to a previously sent message.
    fn replies_expected(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.def().transitions.iter() {
            let replies: Vec<_> = transition
                .out_msgs
                .as_ref()
                .iter()
                .map(|dest| &dest.msg)
                .filter(|msg| msg.is_reply())
                .collect();
            if replies.is_empty() {
                continue;
            }
            if replies.len() > 1 {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(
                                reply.msg_type.span(),
                                error::msgs::MULTIPLE_REPLIES_ERR,
                            )
                        })
                        .collect(),
                );
            }
            if transition.in_msg().is_none() {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(reply.msg_type.span(), error::msgs::INVALID_REPLY_ERR)
                        })
                        .collect(),
                );
            }
        }
        err
    }

    /// A client is any actor with a state that sends at least one message when not handling an
    /// incoming message. Such actors are not allowed to receive any messages which are not replies.
    fn clients_only_receive_replies(&self) -> MaybeErr {
        self.actors()
            .values()
            .filter(|actor| actor.is_client())
            .flat_map(|actor| {
                actor
                    .states()
                    .values()
                    .flat_map(|state| state.methods().values())
            })
            .filter(|method| {
                if let Some(msg) = method.def().in_msg() {
                    !msg.is_reply()
                } else {
                    false
                }
            })
            .map(|transition| {
                syn::Error::new(
                    transition.span(),
                    error::msgs::CLIENT_RECEIVED_NON_REPLY_ERR,
                )
            })
            .collect()
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{
        error::{assert_err, assert_ok},
        parsing::{ActorDef, Dest, NameDef, Protocol, Transition},
    };

    #[test]
    fn all_states_declared_and_used_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.all_states_declared_and_used();

        assert_ok(result);
    }

    #[test]
    fn all_states_declared_and_used_end_not_used_ok() {
        const STATE_NAME: &str = "Init";
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", [STATE_NAME])],
            [Transition::new(
                State::new(STATE_NAME, []),
                Some(Message::new("Activate", false, [])),
                [State::new(STATE_NAME, [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_ok(result);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_out_state_owned_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Init", []), State::new("Next", ["Undeclared"])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_in_state_owned_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", ["Undeclared"]),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_unused_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unused"),
            [ActorDef::new("actor", ["Init", "Extra"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("End", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.all_states_declared_and_used();

        assert_err(result, error::msgs::UNUSED_STATE_ERR);
    }

    #[test]
    fn receivers_and_senders_matched_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.receivers_and_senders_matched();

        assert_ok(result);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_sender_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unbalanced"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                None,
                [State::new("Init", [])],
                [Dest::new(
                    DestinationState::Service(State::new("Init", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_err(result, error::msgs::UNMATCHED_SENDER_ERR);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_receiver_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Unbalanced"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("NotExists", false, [])),
                [State::new("Init", [])],
                [],
            )],
        ))
        .unwrap();

        let result = input.receivers_and_senders_matched();

        assert_err(result, error::msgs::UNMATCHED_RECEIVER_ERR);
    }

    #[test]
    fn no_undeliverable_msgs_ok() {
        let input = ProtocolModel::new(Protocol::minimal()).unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_reply_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Listening", "Client"])],
            [Transition::new(
                State::new("Listening", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Listening", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_service_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Service(State::new("Server", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_owned_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["FileClient", "FileHandle"])],
            [Transition::new(
                State::new("FileClient", ["FileHandle"]),
                None,
                [State::new("FileClient", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("FileHandle", [])),
                    Message::new("FileOp", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_ok(result);
    }

    #[test]
    fn no_undeliverable_msgs_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.no_undeliverable_msgs();

        assert_err(result, error::msgs::UNDELIVERABLE_ERR);
    }

    #[test]
    fn replies_expected_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_ok(result);
    }

    #[test]
    fn replies_expected_invalid_reply_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_err(result, error::msgs::INVALID_REPLY_ERR);
    }

    #[test]
    fn replies_expected_multiple_replies_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "OtherClient", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [
                    Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        Message::new("Msg", true, []),
                    ),
                    Dest::new(
                        DestinationState::Individual(State::new("OtherClient", [])),
                        Message::new("Msg", true, []),
                    ),
                ],
            )],
        ))
        .unwrap();

        let result = input.replies_expected();

        assert_err(result, error::msgs::MULTIPLE_REPLIES_ERR);
    }

    #[test]
    fn clients_only_receive_replies_ok() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ClientReplies"),
            [
                ActorDef::new("client", ["Client", "Waiting"]),
                ActorDef::new("server", ["Init", "Listening"]),
            ],
            [
                Transition::new(
                    State::new("Init", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Listening", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Waiting", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Listening", [])),
                        Message::new("Ping", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Listening", []),
                    Some(Message::new("Ping", false, [])),
                    [State::new("Listening", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Waiting", [])),
                        Message::new("Ping", true, []),
                    )],
                ),
                Transition::new(
                    State::new("Waiting", []),
                    Some(Message::new("Ping", true, [])),
                    [State::new("Client", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.clients_only_receive_replies();

        assert_ok(result);
    }

    #[test]
    fn clients_only_receive_replies_err() {
        let input = ProtocolModel::new(Protocol::new(
            NameDef::new("ClientReplies"),
            [
                ActorDef::new("client", ["Client", "Waiting"]),
                ActorDef::new("server", ["Init", "Listening"]),
            ],
            [
                Transition::new(
                    State::new("Init", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Listening", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Waiting", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Listening", [])),
                        Message::new("Ping", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Listening", []),
                    Some(Message::new("Ping", false, [])),
                    [State::new("Listening", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Waiting", [])),
                        Message::new("Ping", true, []),
                    )],
                ),
                Transition::new(
                    State::new("Waiting", []),
                    Some(Message::new("Ping", true, [])),
                    [State::new("Client", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Client", [])],
                    [],
                ),
            ],
        ))
        .unwrap();

        let result = input.clients_only_receive_replies();

        assert_err(result, error::msgs::CLIENT_RECEIVED_NON_REPLY_ERR);
    }
}
